# Copyright (C) 2025 fontconfig Authors
# SPDX-License-Identifier: HPND

from fctest import FcTest, FcTestFont
from pathlib import Path
from tempfile import TemporaryDirectory
import hashlib
import os
import pytest
import shutil


@pytest.fixture
def fctest():
    return FcTest()


@pytest.fixture
def fcfont():
    return FcTestFont()


@pytest.mark.skipif(not not os.getenv('EXEEXT'), reason='not working on Win32')
def test_sysroot(fctest, fcfont):
    basedir = TemporaryDirectory(prefix='fontconfig.',
                                 suffix='.sysroot')
    cachedir = Path(basedir.name) / Path(fctest.cachedir.name).relative_to('/')
    cachedir.mkdir(parents=True, exist_ok=True)
    fontdir = Path(basedir.name) / Path(fctest.fontdir.name).relative_to('/')
    fontdir.mkdir(parents=True, exist_ok=True)
    configdir = Path(basedir.name) / Path(fctest._conffile.name).parent.relative_to('/')
    configdir.mkdir(parents=True, exist_ok=True)
    fctest.setup()
    fctest.install_font(fcfont.fonts[0], fontdir)
    shutil.copy2(fctest._conffile.name, configdir / Path(fctest._conffile.name).name)
    for ret, stdout, stderr in fctest.run_cache(['-y', basedir.name]):
        assert ret == 0, stderr
    font_files = [fn.name for fn in fontdir.glob('*.pcf')]
    assert len(font_files) == 1, font_files
    cache_files = [c.name for c in cachedir.glob('*cache*')]
    assert len(cache_files) == 1, cache_files

    md5 = hashlib.md5()
    md5.update(fctest.fontdir.name.encode('utf-8'))
    cache_files_based_on_md5 = [c.name for c in cachedir.glob(f'{md5.hexdigest()}*')]
    assert len(cache_files_based_on_md5) == 1, cache_files_based_on_md5



#! /usr/bin/env python3
# Copyright (C) 2026 fontconfig Authors
# SPDX-License-Identifier: HPND

import argparse
import re
import subprocess
import sys
from junit_xml import TestSuite, TestCase


def pickup(fname):
    with open(fname, "r", encoding="utf-8") as f:
        rettype = None
        for l in f.readlines():
            l = l.rstrip()
            m = re.match(r"^([\w][\w \*].*[^;,{}]$)", l)
            if m and m.group(1):
                rettype = m.group(1)
                continue
            m = re.match(r"^(Fc[^ ]*)[\s\w]*\(.*", l)
            if m and m.group(1) not in ["FcCacheDir", "FcCacheSubdir"]:
                if not rettype:
                    raise RuntimeError(f"No return type detected for {m.group(1)}")
                yield (rettype, m.group(1))
            rettype = None


if __name__ == "__main__":
    results = []
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-i",
        "--ignore-implicit-visibility",
        action="store_true",
        help="Ignore implicit symbol visibility",
    )
    parser.add_argument(
        "-f",
        "--ignore-symbol-file",
        type=open,
        help="A file that contains symbol list to ignore implicit visibility",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=argparse.FileType("w"),
        default=sys.stdout,
        help="Output xml file",
    )
    parser.add_argument("library", help="A path to library")
    parser.add_argument("headers", nargs="+", help="A path to header file")

    args = parser.parse_args()
    symbollist = []

    if args.ignore_symbol_file:
        symbollist = [line.rstrip() for line in args.ignore_symbol_file.readlines()]
        symbollist = [item for item in symbollist if not re.match(r"^#.*", item)]

    knownapi = {}
    for fname in args.headers:
        ts = TestCase(
            f"Checking header files {fname}",
            "fontconfig",
            0,
            open(fname).read(),
            allow_multiple_subelements=True,
        )
        for rettype, func in pickup(fname):
            if not re.search(r"FcPrivate|FcPublic", rettype):
                if not args.ignore_implicit_visibility and func not in symbollist:
                    ts.add_failure_info(
                        f"{func} does not have FcPublic nor FcPrivate in {fname}",
                        "",
                        "Missing",
                    )

            knownapi[func] = rettype
        results += [ts]

    res = subprocess.run(
        ["nm", args.library], capture_output=True, text=True, check=True
    )
    if res.returncode != 0:
        raise RuntimeError(f"Unable to decode {args.library}: {res.stderr}")
    ts = TestCase(
        "Checking symbols", "fontconfig", 0, res.stdout, allow_multiple_subelements=True
    )
    for line in res.stdout.splitlines():
        m = re.search(r"\sT\s(.*)$", line)
        if m and m.group(1):
            if m.group(1) not in knownapi and m.group(1) not in symbollist:
                ts.add_failure_info(
                    f"{m.group(1)} does not have FcPublic nor FcPrivate", "", "Missing"
                )
    results += [ts]
    args.output.write(TestSuite.to_xml_string([TestSuite("API checking", results)]))
    args.output.close()



#! /usr/bin/env python3
# Copyright (C) 2025 fontconfig Authors
# SPDX-License-Identifier: HPND

import sys
import re
import argparse


def gen_header():
    return '''/* Copyright (C) 2025 fontconfig Authors */
/* SPDX-License-Identifier: HPND */

'''


def parse_list(lfile):
    with open(lfile) as f:
        lines = []
        for s in f:
            x = s.split()
            if not s.startswith('#') and len(x) > 0:
                lines.append(x)
    return sorted(lines, key=lambda a: a[0].lower())


def parse_fcobjh(bfile):
    with open(bfile) as f:
        s = f.read()
        s = re.sub(re.compile(r'//.*?$|/\*.*?\*/|\'(?:\\.|[^\\\'])*\'|"(?:\\.|[^\\"])*"',
                              re.DOTALL | re.MULTILINE),
                   lambda m: " " if m.group(0).startswith('/') else s,
                   s)
        l = [ss.strip().replace('FC_OBJECT (', '').replace(')','').split() for ss in s.splitlines()]
        ll = [x for x in l if x]
        # 0 is reserved for FC_INVALID_OBJECT
        ret = dict([('FC_' + ll[n][0].strip(','), n + 1) for n in range(0, len(ll))])
        rret = dict([(n + 1, 'FC_' + ll[n][0].strip(',')) for n in range(0, len(ll))])
        ret['FC_INVALID_OBJECT'] = 0
        rret[0] = 'FC_INVALID_OBJECT'
        return ret, rret


def gen_constsym(sym_list, max_sym):
    ret = ['typedef struct _FcConstIndex {',
           '    FcObject object;',
           '    int      idx_obj;',
           '    int      idx_variant;',
           '} FcConstIndex;',
           '',
           'typedef struct _FcConstSymbolMap {',
           '    const FcChar8 *name;',
           f'    FcConstIndex   values[{max_sym+1}];',
           '} FcConstSymbolMap;',
           '',
           'static const FcConstSymbolMap _FcBaseConstantSymbols[] = {']
    for k in sorted(sym_list.keys()):
        a = sym_list[k]
        ret.append('    {')
        ret.append(f'        (const FcChar8 *) "{k}",')
        ret.append('        {')
        for v in a:
            ret.append(f'            {{ {v[0]}_OBJECT, {v[1]}, {v[2]} }},')
        ret.append('            { FC_INVALID_OBJECT, 0, 0 },')
        ret.append('        },')
        ret.append('    },')
    ret.append('};')
    ret.append('')
    ret.append('#define NUM_FC_CONST_SYMBOLS (sizeof (_FcBaseConstantSymbols) / sizeof (_FcBaseConstantSymbols[0]))')
    ret.append('')
    return ret


def gen_baseconstobj_body(objs, enum_list):
    max_objs = 0
    ret = []
    for i, a in enumerate(objs):
        if not a:
            ret.append('    {{{ NULL, NULL, 0 }}}    /* ' + enum_list[i] + ' */,')
        else:
            ret.append('    {{')
            n = 0
            for v in a:
                ret.append('        { (const FcChar8 *) "' + v[0] + '", ' + enum_list[i] + ', ' + v[1] + '},')
                n += 1
            ret.append('        { NULL, NULL, 0 },')
            max_objs = max([n + 1, max_objs])
            ret.append('    }},')
    return ret, max_objs


def gen_decl_baseconstobj(max_objs):
    return ['typedef struct _fcConstObjects {',
            f'    FcConstant values[%d];' % max_objs,
            '} FcConstantObjects;',
            '',
            'static const FcConstantObjects _FcBaseConstantObjects[FC_MAX_BASE_OBJECT+1] = {'
            ]


def gen_baseconstobj(objs, enum_list):
    body, max_objs = gen_baseconstobj_body(objs, enum_list)
    decl = gen_decl_baseconstobj(max_objs)
    return decl + body + ['};', '', '#define NUM_FC_CONST_OBJS (sizeof (_FcBaseConstantObjects) / sizeof (_FcBaseConstantObjects[0]))', '']


def gen_body(lfile, bfile):
    const_list = parse_list(lfile)
    enum_list, reverse_enum_list = parse_fcobjh(bfile)
    const_list_by_objs = [None] * len(enum_list)
    sym_list_by_const = {}
    max_sym = 0
    for i, a in enumerate(const_list):
        idx = enum_list[a[1]]
        if not isinstance(const_list_by_objs[idx], list):
            const_list_by_objs[idx] = []
        const_list_by_objs[idx].append([a[0], a[2]])
        if a[0] not in sym_list_by_const:
            sym_list_by_const[a[0]] = []
        sym_list_by_const[a[0]].append([a[1], idx, len(const_list_by_objs[idx]) - 1])
        max_sym = max([len(sym_list_by_const[a[0]]), max_sym])
    sym = gen_constsym(sym_list_by_const, max_sym)
    objs = gen_baseconstobj(const_list_by_objs, reverse_enum_list)

    return '\n'.join(sym) + '\n'+ '\n'.join(objs)


def gen_test_body(lfile, bfile):
    ret = []
    const_list = parse_list(lfile)
    enum_list, reverse_enum_list = parse_fcobjh(bfile)
    ret.append('#include <stdio.h>')
    ret.append('#include "fontconfig/fontconfig.h"')
    ret.append('')
    ret.append('int test (void) {')
    ret.append('    int ret = 0;')
    ret.append('    const FcConstant *c;')
    for i, a in enumerate(const_list):
        ret.append(f'    c = FcNameGetConstantFor ((const FcChar8 *)"{a[0]}", {a[1]});')
        ret.append(f'    if (!c || c->value != {a[2]}) {{')
        ret.append(f'        fprintf (stderr, "failed: (%s, %s)\\n", "{a[0]}", _FC_STRINGIFY({a[1]}));')
        ret.append('        ret++;')
        ret.append('    }')
    ret.append('    return ret;')
    ret.append('}')
    ret.append('')
    ret.append('int main(void) { return test(); }')
    ret.append('')
    return '\n'.join(ret)


def main():
    parser = argparse.ArgumentParser(
        description='fcconst.h generator',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument('-o', '--output',
                        type=argparse.FileType('w'),
                        default='-',
                        help='Output file')
    parser.add_argument('-t', '--test',
                        action='store_true',
                        help='Generate test case')
    parser.add_argument('list',
                        help='list file for constant names')
    parser.add_argument('header',
                        help='fcobjs.h file')

    args = parser.parse_args()
    with args.output:
        args.output.write(gen_header())
        if args.test:
            args.output.write(gen_test_body(args.list, args.header))
        else:
            args.output.write(gen_body(args.list, args.header))

    sys.exit(0)


if __name__ == '__main__':
    main()



#! /usr/bin/env python3
# Copyright (C) 2026 fontconfig Authors
# SPDX-License-Identifier: HPND

import argparse
import re
import sys
from enum import Enum
from junit_xml import TestSuite, TestCase


class ResultContext(Enum):
    Unknown = 0
    Summary = 1
    Added = 2
    Removed = 3


def is_error(ctx):
    return ctx != ResultContext.Summary


def output(head, lines, ctx):
    t = TestCase(
        head, "fontconfig", 0, "\n".join(lines), allow_multiple_subelements=True
    )
    if is_error(ctx):
        for line in lines:
            t.add_failure_info(head, line, ctx)

    return [t]


def abidiff2junit(result):
    results = []

    context = ResultContext.Unknown
    head = ""
    stdout = []
    unknown = []
    for line in result.splitlines():
        if m := re.match(r"(.*) changes summary: (.*)", line):
            ctx = ResultContext.Summary
            h = m.group(1) + " changes summary"
            if context != ResultContext.Unknown and (ctx != context or h != head):
                results += output(head, stdout, context)
            context = ctx
            head = m.group(1) + " changes summary"
            stdout = [m.group(2)]
        elif m := re.match(r"(\d+) (Removed|Added)\s(.*):$", line):
            ctx = (
                ResultContext.Added if m.group(2) == "Added" else ResultContext.Removed
            )
            if context != ResultContext.Unknown and ctx != context:
                results += output(head, stdout, context)
            context = ctx
            head = m.group(2) + " " + m.group(3)
            stdout = []
        elif not line.strip():
            continue
        elif context == ResultContext.Added or context == ResultContext.Removed:
            stdout += [line]
        else:
            unknown += [line]
    if context != ResultContext.Unknown:
        results += output(head, stdout, context)
    results += output("Unknown response", unknown, ResultContext.Unknown)
    return TestSuite("The result of abidiff", results)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-o",
        "--output",
        type=argparse.FileType("w"),
        default=sys.stdout,
        help="Outout xml file",
    )
    parser.add_argument(
        "RESULT",
        type=argparse.FileType("r"),
        default=sys.stdin,
        help="Result from abidiff",
    )
    args = parser.parse_args(args=sys.argv[1:])
    res = abidiff2junit(args.RESULT.read())
    args.output.write(TestSuite.to_xml_string([res]))
    args.output.close()



# Copyright (C) 2025 fontconfig Authors
# SPDX-License-Identifier: HPND

from fctest import FcTest, FcTestFont
from pathlib import Path
from tempfile import TemporaryDirectory, NamedTemporaryFile
import os
import pytest
import time
import types


@pytest.fixture
def fctest():
    return FcTest()


@pytest.fixture
def fcfont():
    return FcTestFont()


@pytest.mark.skipif(not not os.getenv('EXEEXT'), reason='not working on Win32')
def test_crbug1004254(fctest, fcfont):
    builddir = Path(fctest.builddir)
    def custom_config(self):
        return f'''
<fontconfig>
  <dir>{builddir.resolve()}/testfonts</dir>
  <cachedir>{fctest.cachedir.name}</cachedir>
</fontconfig>'''

    fctest.config = types.MethodType(custom_config, fctest)
    fctest.setup()

    testexe = Path(fctest.builddir) / 'test' / ('test-crbug1004254' + fctest._exeext)
    if not testexe.exists():
        testexe = Path(fctest.builddir) / 'test' / ('test_crbug1004254' + fctest._exeext)
        if not testexe.exists():
            raise RuntimeError('No test case for crbug1004254')

    for ret, stdout, stderr in fctest.run(testexe):
        assert ret == 0, stderr
        fctest.logger.info(stdout)
        fctest.logger.info(stderr)
        fctest.logger.info(fctest.config())



#! /usr/bin/env python3
# Copyright (C) 2026 fontconfig Authors
# SPDX-License-Identifier: HPND

import argparse
import os
import pathlib
import platform
import re
import shutil
import subprocess
import sys
import tempfile
import time
import traceback
from typing import List, Dict, Optional

# PyYAML is loaded after venv setup (installed in venv, not required system-wide)
try:
    import yaml
except ImportError:
    yaml = None


# Global log file descriptor
_global_logfd = None


def log(message, end='\n'):
    """Write to both stdout and log file"""
    print(message, end=end, flush=True)
    if _global_logfd and not _global_logfd.closed:
        try:
            _global_logfd.write(message + end)
            _global_logfd.flush()
        except:
            pass


def get_venv_paths(venv_path: pathlib.Path) -> tuple:
    """Get venv bin directory and executable paths based on platform"""
    # MSYS2 uses bin/ with .exe files, pure Windows uses Scripts/
    scripts_dir = venv_path / "Scripts"
    bin_dir = venv_path / "bin"

    # Determine venv bin directory
    venv_bin = scripts_dir if scripts_dir.exists() else \
               bin_dir if bin_dir.exists() else \
               scripts_dir if sys.platform == "win32" else bin_dir

    # Determine executable suffix
    exe_suffix = ".exe" if (venv_bin / "python.exe").exists() or \
                          (not venv_bin.exists() and sys.platform == "win32") else ""

    return venv_bin, venv_bin / f"python{exe_suffix}", venv_bin / f"pip{exe_suffix}"


class Build:
    def __init__(self, args):
        self.args = args
        self.cidir = pathlib.Path(__file__).parent
        self.srcdir = pathlib.Path.cwd()
        self.task = None
        self.config = {}

        self._setup_environment()
        self._load_config()

    def _log(self, message, end='\n'):
        """Write to both stdout and log file"""
        log(message, end=end)

    def _load_fcenv(self):
        """Load environment variables from fcenv file if it exists"""
        fcenv_path = self.cidir / "fcenv"
        if not fcenv_path.exists():
            return

        log(f">>> Loading environment from {fcenv_path}")

        var_pattern = re.compile(r'^([A-Z_][A-Z0-9_]*)=(.*)$')

        try:
            with open(fcenv_path, 'r') as f:
                for line in f:
                    line = line.strip()
                    if not line or line.startswith('#'):
                        continue

                    if match := var_pattern.match(line):
                        key, value = match.groups()
                        # Remove surrounding quotes if present
                        if (value.startswith('"') and value.endswith('"')) or \
                           (value.startswith("'") and value.endswith("'")):
                            value = value[1:-1]
                        os.environ[key] = value
                        log(f"  {key}={value}")
        except Exception as e:
            log(f"Warning: Failed to load fcenv: {e}")

    def _setup_environment(self):
        """Setup environment variables and detect OS"""
        self._load_fcenv()

        # Set environment variable defaults
        os.environ.setdefault("MAKE", "make")
        os.environ.setdefault("BUILD_ID", f"fontconfig-{os.getpid()}")
        os.environ.setdefault("PREFIX", str(self.srcdir / "prefix"))
        os.environ.setdefault("BUILDDIR", str(self.srcdir / "build"))
        os.environ.setdefault("CI_MERGE_REQUEST_PROJECT_URL", "https://gitlab.freedesktop.org/fontconfig/fontconfig")
        os.environ.setdefault("CI_COMMIT_REF_NAME", "main")
        os.environ.setdefault("BUILDLOG", str(pathlib.Path(os.environ["BUILDDIR"]) / "fc-build.log"))

        self.builddir = pathlib.Path(os.environ["BUILDDIR"])
        self.prefix = pathlib.Path(os.environ["PREFIX"])
        self.buildlog = pathlib.Path(os.environ["BUILDLOG"])

        if not os.environ.get("FC_DISTRO_NAME"):
            self._detect_distro()

        self.distro_name = os.environ.get("FC_DISTRO_NAME", "")

    def _detect_distro(self):
        """Detect distribution from platform or /etc/os-release"""
        system = platform.system()

        if system == 'Darwin':
            os.environ["FC_DISTRO_NAME"] = "darwin"
        elif system == 'Windows':
            os.environ["FC_DISTRO_NAME"] = "windows"
        else:
            # Try Linux /etc/os-release
            try:
                with open("/etc/os-release") as f:
                    for line in f:
                        if line.startswith("ID="):
                            os.environ["FC_DISTRO_NAME"] = line.split("=")[1].strip().strip('"')
                            return
            except FileNotFoundError:
                pass

            log("***\n*** Unable to detect OS. cross-compiling may not work. "
                "Please consider setting FC_DISTRO_NAME\n***")
            time.sleep(3)

    def _load_config(self):
        """Load build configuration from YAML"""
        config_file = self.cidir / "build.yaml"
        try:
            with open(config_file, 'r') as f:
                self.config = yaml.safe_load(f) or {}
        except Exception as e:
            log(f"Warning: Failed to load build config: {e}")
            self.config = {}

    def _set_env_vars(self, env_vars: Dict[str, str], context: str):
        """Set environment variables from config"""
        self._log(f"\n>>> Setting {context} environment variables")
        for key, value in env_vars.items():
            expanded = os.path.expandvars(value)
            os.environ[key] = expanded
            self._log(f"  {key}={expanded}")

    def _setup_certificates(self):
        """Setup SSL certificate paths using certifi"""
        try:
            result = subprocess.run(
                [sys.executable, "-m", "certifi"],
                capture_output=True, text=True, check=True
            )
            if cert_path := result.stdout.strip():
                os.environ["SSL_CERT_FILE"] = os.environ["REQUESTS_CA_BUNDLE"] = cert_path
                self._log(f"\n>>> Setting certificate paths:\n  SSL_CERT_FILE={cert_path}\n  REQUESTS_CA_BUNDLE={cert_path}")
        except Exception as e:
            self._log(f"\nWarning: Failed to setup certificates: {e}")

    def _stream_subprocess(self, args: List[str], shell: bool = False,
                          cwd: Optional[pathlib.Path] = None) -> subprocess.CompletedProcess:
        """Run subprocess and stream output to log"""
        proc = subprocess.Popen(
            args,
            stderr=subprocess.STDOUT,
            stdout=subprocess.PIPE,
            text=True,
            cwd=cwd or self.srcdir,
            shell=shell,
            env=os.environ.copy()
        )

        for line in proc.stdout:
            self._log(line, end='')

        proc.wait()
        return subprocess.CompletedProcess(args, proc.returncode)

    def _run_shell_script(self, script: str, shell: str = 'sh',
                         description: str = "", ignore_errors: bool = False) -> bool:
        """Run a shell script and log output"""
        if not script.strip():
            return True

        if description:
            self._log(f"\n>>> {description}")

        result = self._stream_subprocess([shell, '-c', script], shell=False)

        if result.returncode != 0 and not ignore_errors:
            error_msg = f"{description} failed" if description else "Script failed"
            raise subprocess.CalledProcessError(result.returncode, f"{shell} -c <script>", output=error_msg)

        return result.returncode == 0

    def _run_phase_script(self, phase_cfg: Dict, context: str, ignore_errors: bool = False):
        """Run a phase configuration (env + script)"""
        if not phase_cfg:
            return

        # Run script first (may create directories/install dependencies)
        if script := phase_cfg.get('script'):
            self._run_shell_script(script, phase_cfg.get('shell', 'sh'),
                                  f"Running {context} script", ignore_errors)

        # Then set environment variables (may reference things created by script)
        if env_vars := phase_cfg.get('env'):
            self._set_env_vars(env_vars, context)

    def _run_distro_script(self, phase: str):
        """Run distro-specific script for a given phase"""
        if not self.distro_name:
            return

        distro_cfg = self.config.get(self.distro_name, {})
        phase_cfg = distro_cfg.get(phase)

        self._run_phase_script(phase_cfg, f"{self.distro_name} {phase}")

        # Handle certificate setup for macOS and similar platforms
        if phase_cfg and phase_cfg.get('needs_cert_setup'):
            self._setup_certificates()

    def _run_feature_setup(self, feature: str):
        """Run feature-specific setup from config"""
        features = self.config.get('common', {}).get('features', {})
        self._run_phase_script(features.get(feature), f"{feature} feature", ignore_errors=True)

    def _run_command(self, args: List[str], task: str, cwd: Optional[pathlib.Path] = None):
        """Run a command and log output"""
        self.task = task
        self._log(f"\n>>> Running: {task}")
        self._log(f">>> Command: {' '.join(args)}")

        if cwd is None:
            cwd = self.builddir if getattr(self, '_in_builddir', False) else None
        result = self._stream_subprocess(args, cwd=cwd)

        if result.returncode != 0:
            raise subprocess.CalledProcessError(result.returncode, args)

    def _clean_paths(self, *paths: pathlib.Path):
        """Remove paths if they exist"""
        for path in paths:
            if path.exists():
                shutil.rmtree(path)

    def _is_cross(self) -> bool:
        """Check if this is a cross-compile build"""
        return os.environ.get("FC_BUILD_PLATFORM") in {"mingw", "android"}

    def prepare(self):
        raise NotImplementedError

    def compile(self):
        raise NotImplementedError

    def install(self):
        raise NotImplementedError

    def test(self):
        raise NotImplementedError

    def distcheck(self):
        raise NotImplementedError


class BuildMeson(Build):
    def __init__(self, args):
        super().__init__(args)
        self.buildopt = []
        self.subprojectname = ""

    def _meson_cmd(self, *args):
        """Run meson command with standard options"""
        return ["meson", *args, "-C", str(self.builddir)]

    def _handle_features(self):
        """Handle enabled/disabled features"""
        if self.args.enable_feature:
            for feature in self.args.enable_feature:
                self.buildopt.append(f"-D{self.subprojectname}{feature}=enabled")
                self._run_feature_setup(feature)

        if self.args.disable_feature:
            self.buildopt.extend(
                f"-D{self.subprojectname}{f}=disabled"
                for f in self.args.disable_feature
            )

    def _handle_subproject_build(self):
        """Handle subproject build setup"""
        if not self.args.subproject_build:
            return

        self.buildopt.append("--force-fallback-for=fontconfig")

        subproject_dir = self.srcdir / "fc-ci-meson-subproject"
        self._clean_paths(subproject_dir)

        # Clone fc-ci-meson-subproject
        self._run_command([
            "git", "clone",
            "https://gitlab.freedesktop.org/fontconfig/fontconfig-ci/fc-ci-meson-subproject.git"
        ], "git clone", cwd=self.srcdir)

        # Clone fontconfig into subprojects
        subprojects_dir = subproject_dir / "subprojects"
        self._run_command([
            "git", "clone",
            f"{os.environ['CI_MERGE_REQUEST_PROJECT_URL']}.git"
        ], "git clone fontconfig", cwd=subprojects_dir)

        # Handle merge request
        if iid := os.environ.get("CI_MERGE_REQUEST_IID"):
            fontconfig_dir = subprojects_dir / "fontconfig"
            self._run_command([
                "git", "fetch", "origin",
                f"merge-requests/{iid}/head:mr-{iid}"
            ], "git fetch MR", cwd=fontconfig_dir)
            self._run_command([
                "git", "switch", f"mr-{iid}"
            ], "git switch MR", cwd=fontconfig_dir)

        # Update srcdir to point to subproject directory for subsequent meson commands
        self.srcdir = subproject_dir
        self.subprojectname = "fontconfig:"

    def _build_options(self):
        """Build meson options"""
        self._handle_features()
        self._handle_subproject_build()

        if self.args.optimization:
            self.buildopt.append(f"--optimization={self.args.optimization}")

        if self.args.xmlbackend:
            self.buildopt.append(f"-D{self.subprojectname}xml-backend={self.args.xmlbackend}")

        if self._is_cross() and self.args.arch:
            self.buildopt.extend([
                "--cross-file", str(self.cidir / f"{self.args.arch}.txt")
            ])
            self._run_distro_script('cross')

        self.buildopt.append(f"--default-library={self.args.buildtype}")

        if self.args.sanitize:
            self.buildopt.extend([
                f"-Db_sanitize={self.args.sanitize}",
                "-Db_lundef=false"
            ])

    def prepare(self):
        """Prepare meson build"""
        self._build_options()

        if not self.args.no_clean:
            self._clean_paths(self.builddir, self.prefix)

        self._run_command([
            "meson", "setup",
            f"--prefix={self.prefix}",
            f"-D{self.subprojectname}nls=enabled",
            f"-D{self.subprojectname}cache-build=disabled",
            f"-D{self.subprojectname}iconv=enabled",
            *self.buildopt,
            str(self.builddir)
        ], "meson setup")

    def compile(self):
        self._run_command(self._meson_cmd("compile", "-v"), "meson compile")

    def install(self):
        self._run_command(self._meson_cmd("install"), "meson install")

    def test(self):
        self._run_command(self._meson_cmd("test", "-v"), "meson test")

    def distcheck(self):
        self._run_command(self._meson_cmd("dist"), "meson dist")


class BuildAutotools(Build):
    def __init__(self, args):
        super().__init__(args)
        self.buildopt = []
        self.make = os.environ.get("MAKE", "make")

        if args.subproject_build:
            raise RuntimeError("Subproject build not supported in autotools")

    def _make_cmd(self, *targets):
        """Build make command with targets"""
        return [self.make, *targets, "V=1"]

    def _build_options(self):
        """Build autotools configure options"""
        if self.args.enable_feature:
            self.buildopt.extend(f"--enable-{f}" for f in self.args.enable_feature)

        if self.args.disable_feature:
            self.buildopt.extend(f"--disable-{f}" for f in self.args.disable_feature)

        # XML backend options
        if self.args.xmlbackend == "expat":
            self.buildopt.append("--disable-libxml2")
        elif self.args.xmlbackend == "libxml2":
            self.buildopt.append("--enable-libxml2")

        # Build type options
        self.buildopt.extend({
            "shared": ["--enable-shared", "--disable-static"],
            "static": ["--disable-shared", "--enable-static"],
            "both": ["--enable-shared", "--enable-static"]
        }.get(self.args.buildtype, []))

        if self._is_cross() and self.args.arch:
            self.buildopt.append(f"--host={self.args.arch}")
            self._run_distro_script('cross')

    def prepare(self):
        """Prepare autotools build"""
        self._build_options()

        if not self.args.no_clean:
            self._clean_paths(self.builddir, self.prefix)

        self.builddir.mkdir(parents=True, exist_ok=True)
        self.prefix.mkdir(parents=True, exist_ok=True)

        os.chdir(self.builddir)
        self._in_builddir = True

        self._run_command([
            str(self.srcdir / "autogen.sh"),
            f"--prefix={self.prefix}",
            "--disable-cache-build",
            *self.buildopt
        ], "autogen.sh")

    def compile(self):
        self._run_command(self._make_cmd(), "make")

    def install(self):
        self._run_command(self._make_cmd("install"), "make install")

    def test(self):
        self._run_command(self._make_cmd("check"), "make check")

    def distcheck(self):
        self._run_command(self._make_cmd("distcheck"), "make distcheck")


def ensure_venv(skip_venv=False):
    """Ensure build.py is running in a virtual environment"""
    if skip_venv or hasattr(sys, 'real_prefix') or sys.base_prefix != sys.prefix:
        return

    venv_path = pathlib.Path.cwd() / ".venv"
    needs_setup = not venv_path.exists()

    # Create venv if needed
    if needs_setup:
        log(">>> Creating Python virtual environment")
        if subprocess.run([sys.executable, "-m", "venv", str(venv_path)]).returncode != 0:
            sys.exit("Error: venv creation failed")

    venv_bin, venv_python, venv_pip = get_venv_paths(venv_path)

    # Install PyYAML if venv was just created
    if needs_setup:
        log(">>> Installing PyYAML in virtual environment")
        if subprocess.run([str(venv_pip), "install", "PyYAML"]).returncode != 0:
            sys.exit("Error: PyYAML installation failed")

    # Activate venv and re-execute
    os.environ.update({"PATH": f"{venv_bin}:{os.environ['PATH']}", "VIRTUAL_ENV": str(venv_path)})
    log(">>> Re-executing build.py in virtual environment")
    sys.exit(subprocess.run([str(venv_python)] + sys.argv).returncode)


def main():
    global _global_logfd

    # Create log file early so all output is captured
    temp_logfile = tempfile.NamedTemporaryFile(
        mode='w', prefix='fc-build-', suffix='.log',
        delete=False
    )
    _global_logfd = temp_logfile

    parser = argparse.ArgumentParser(
        description="Build fontconfig using meson or autotools",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""
Environment variables:
  FC_BUILD_PLATFORM  Set a platform for cross-compiling (mingw or android)
  FC_DISTRO_NAME     Distribution name for distro-specific setup
"""
    )
    parser.add_argument("-a", "--arch", help="Target architecture to cross-compile")
    parser.add_argument("-A", "--sanitize", help="Enable the address sanitizer (meson only)")
    parser.add_argument("-c", "--distcheck", action="store_true", help="Run distcheck")
    parser.add_argument("-C", "--no-test", action="store_true", help="Do not run unit tests")
    parser.add_argument("-e", "--enable-feature", action="append",
                       help="Enable specific features (can be used multiple times)")
    parser.add_argument("-d", "--disable-feature", action="append",
                       help="Disable specific features (can be used multiple times)")
    parser.add_argument("-I", "--no-install", action="store_true", help="Do not run installation")
    parser.add_argument("-N", "--no-clean", action="store_true", help="Do not clean build directory")
    parser.add_argument("-O", "--optimization", help="Optimization level to build")
    parser.add_argument("-V", "--no-venv", action="store_true", help="Do not use virtual environment")
    parser.add_argument("-s", "--buildsys", choices=["meson", "autotools"], default="meson",
                       help="Use the buildsys to build")
    parser.add_argument("-S", "--subproject-build", action="store_true",
                       help="Run the sub-project build (meson only)")
    parser.add_argument("-t", "--buildtype", choices=["shared", "static", "both"], default="shared",
                       help="Build type: shared, static, or both")
    parser.add_argument("-X", "--xmlbackend", choices=["expat", "libxml2"], default="expat",
                       help="XML parser library to use")

    args = parser.parse_args()

    if args.sanitize and not args.optimization:
        args.optimization = "g"

    # Ensure venv after parsing args (allows --help without venv setup)
    ensure_venv(skip_venv=args.no_venv)

    # Install packages from build.yaml (only when using venv)
    if not args.no_venv:
        try:
            with open(pathlib.Path(__file__).parent / "build.yaml", 'r') as f:
                config = yaml.safe_load(f)

            if packages := config.get('common', {}).get('packages', []):
                venv_pip = str(get_venv_paths(pathlib.Path.cwd() / ".venv")[2])
                log(">>> Installing packages from build.yaml")
                if subprocess.run([venv_pip, "install"] + packages).returncode != 0:
                    sys.exit("Error: Package installation failed")
        except FileNotFoundError:
            pass
        except Exception as e:
            log(f"Warning: Failed to install packages from build.yaml: {e}")

    try:
        build = BuildMeson(args) if args.buildsys == "meson" else BuildAutotools(args)

        build._log(f"Build System: {args.buildsys}\n")

        build._log("Environment variables:")
        for key, value in sorted(os.environ.items()):
            build._log(f"{key}={value}")
        build._log("")

        build._run_distro_script('setup')
        build._run_distro_script('install')

        build.prepare()
        build.compile()

        if not args.no_install:
            build.install()

        if not args.no_test:
            build.test()

        if args.distcheck:
            build.distcheck()

        return 0

    except Exception as e:
        # Get task name if build object exists
        task = getattr(locals().get('build'), 'task', None)
        error_prefix = f"\nAborting from \"{task}\" with error: " if task else "\nBuild failed: "
        error_msg = f"{error_prefix}{e}\n{traceback.format_exc()}"

        print(error_msg, file=sys.stderr, end='')
        log(error_msg)
        return 1
    finally:
        # Close log file and move to final location
        if _global_logfd and not _global_logfd.closed:
            temp_log_path = _global_logfd.name
            _global_logfd.close()

            try:
                buildlog = pathlib.Path(os.environ.get("BUILDLOG", "build/fc-build.log"))
                buildlog.parent.mkdir(parents=True, exist_ok=True)
                shutil.move(temp_log_path, buildlog)
            except Exception:
                pass


if __name__ == "__main__":
    sys.exit(main())



# Copyright (C) 2025 fontconfig Authors
# SPDX-License-Identifier: HPND

from fctest import FcTest, FcTestFont
from pathlib import Path
from tempfile import TemporaryDirectory, NamedTemporaryFile
import os
import pytest
import time
import types


@pytest.fixture
def fctest():
    return FcTest()


@pytest.fixture
def fcfont():
    return FcTestFont()


def test_basic(fctest, fcfont):
    fctest.setup()
    fctest.install_font(fcfont.fonts, '.')
    l = []
    for ret, stdout, stderr in fctest.run_list(['-', 'family', 'pixelsize']):
        assert ret == 0, stderr
        l += sorted(stdout.splitlines())
    l += ['=']
    for ret, stdout, stderr in fctest.run_list(['-', 'family', 'pixelsize']):
        assert ret == 0, stderr
        l += sorted(stdout.splitlines())
    l += ['=']
    for ret, stdout, stderr in fctest.run_list(['-', 'family', 'pixelsize']):
        assert ret == 0, stderr
        l += sorted(stdout.splitlines())
    l += ['']
    out = '\n'.join(l)
    with open(Path(fctest.builddir) / 'test' / 'out.expected') as f:
        out_expected = f.read()
        assert out == out_expected


def test_subdir(fctest, fcfont):
    fctest.setup()
    fctest.install_font(fcfont.fonts, 'a')
    for ret, stdout, stderr in fctest.run_cache([fctest.fontdir.name]):
        assert ret == 0, stderr
    l = []
    for ret, stdout, stderr in fctest.run_list(['-', 'family', 'pixelsize']):
        assert ret == 0, stderr
        l += sorted(stdout.splitlines())
    l += ['=']
    for ret, stdout, stderr in fctest.run_list(['-', 'family', 'pixelsize']):
        assert ret == 0, stderr
        l += sorted(stdout.splitlines())
    l += ['=']
    for ret, stdout, stderr in fctest.run_list(['-', 'family', 'pixelsize']):
        assert ret == 0, stderr
        l += sorted(stdout.splitlines())
    l += ['']
    out = '\n'.join(l)
    with open(Path(fctest.builddir) / 'test' / 'out.expected') as f:
        out_expected = f.read()
        assert out == out_expected

def test_subdir_with_cache(fctest, fcfont):
    fctest.setup()
    fctest.install_font(fcfont.fonts, 'a')
    for ret, stdout, stderr in fctest.run_cache([str(Path(fctest.fontdir.name) / 'a')]):
        assert ret == 0, stderr
    l = []
    for ret, stdout, stderr in fctest.run_list(['-', 'family', 'pixelsize']):
        assert ret == 0, stderr
        l += sorted(stdout.splitlines())
    l += ['=']
    for ret, stdout, stderr in fctest.run_list(['-', 'family', 'pixelsize']):
        assert ret == 0, stderr
        l += sorted(stdout.splitlines())
    l += ['=']
    for ret, stdout, stderr in fctest.run_list(['-', 'family', 'pixelsize']):
        assert ret == 0, stderr
        l += sorted(stdout.splitlines())
    l += ['']
    out = '\n'.join(l)
    with open(Path(fctest.builddir) / 'test' / 'out.expected') as f:
        out_expected = f.read()
        assert out == out_expected


def test_with_dotfiles(fctest, fcfont):
    fctest.setup()
    fctest.install_font(fcfont.fonts, '.')
    for f in Path(fctest.fontdir.name).glob('*.pcf'):
        f.rename(f.parent / ('.' + f.name))
    for ret, stdout, stderr in fctest.run_cache([fctest.fontdir.name]):
        assert ret == 0, stderr
    l = []
    for ret, stdout, stderr in fctest.run_list(['-', 'family', 'pixelsize']):
        assert ret == 0, stderr
        l += sorted(stdout.splitlines())
    l += ['=']
    for ret, stdout, stderr in fctest.run_list(['-', 'family', 'pixelsize']):
        assert ret == 0, stderr
        l += sorted(stdout.splitlines())
    l += ['=']
    for ret, stdout, stderr in fctest.run_list(['-', 'family', 'pixelsize']):
        assert ret == 0, stderr
        l += sorted(stdout.splitlines())
    l += ['']
    out = '\n'.join(l)
    with open(Path(fctest.builddir) / 'test' / 'out.expected') as f:
        out_expected = f.read()
        assert out == out_expected


def test_with_dotdir(fctest, fcfont):
    fctest.setup()
    fctest.install_font(fcfont.fonts, '.a')
    for ret, stdout, stderr in fctest.run_cache([fctest.fontdir.name]):
        assert ret == 0, stderr
    l = []
    for ret, stdout, stderr in fctest.run_list(['-', 'family', 'pixelsize']):
        assert ret == 0, stderr
        l += sorted(stdout.splitlines())
    l += ['=']
    for ret, stdout, stderr in fctest.run_list(['-', 'family', 'pixelsize']):
        assert ret == 0, stderr
        l += sorted(stdout.splitlines())
    l += ['=']
    for ret, stdout, stderr in fctest.run_list(['-', 'family', 'pixelsize']):
        assert ret == 0, stderr
        l += sorted(stdout.splitlines())
    l += ['']
    out = '\n'.join(l)
    with open(Path(fctest.builddir) / 'test' / 'out.expected') as f:
        out_expected = f.read()
        assert out == out_expected


def test_with_complicated_dir_structure(fctest, fcfont):
    fctest.setup()
    fctest.install_font(fcfont.fonts[0], Path('a') / 'a')
    fctest.install_font(fcfont.fonts[1], Path('b') / 'b')
    l = []
    for ret, stdout, stderr in fctest.run_list(['-', 'family', 'pixelsize']):
        assert ret == 0, stderr
        l += sorted(stdout.splitlines())
    l += ['=']
    for ret, stdout, stderr in fctest.run_list(['-', 'family', 'pixelsize']):
        assert ret == 0, stderr
        l += sorted(stdout.splitlines())
    l += ['=']
    for ret, stdout, stderr in fctest.run_list(['-', 'family', 'pixelsize']):
        assert ret == 0, stderr
        l += sorted(stdout.splitlines())
    l += ['']
    out = '\n'.join(l)
    with open(Path(fctest.builddir) / 'test' / 'out.expected') as f:
        out_expected = f.read()
        assert out == out_expected


def test_subdir_with_out_of_date_cache(fctest, fcfont):
    fctest.setup()
    fctest.install_font([], 'a')
    for ret, stdout, stderr in fctest.run_cache([str(Path(fctest.fontdir.name) / 'a')]):
        assert ret == 0, stderr
    time.sleep(1)
    fctest.install_font(fcfont.fonts, 'a')
    l = []
    for ret, stdout, stderr in fctest.run_list(['-', 'family', 'pixelsize']):
        assert ret == 0, stderr
        l += sorted(stdout.splitlines())
    l += ['=']
    for ret, stdout, stderr in fctest.run_list(['-', 'family', 'pixelsize']):
        assert ret == 0, stderr
        l += sorted(stdout.splitlines())
    l += ['=']
    for ret, stdout, stderr in fctest.run_list(['-', 'family', 'pixelsize']):
        assert ret == 0, stderr
        l += sorted(stdout.splitlines())
    l += ['']
    out = '\n'.join(l)
    with open(Path(fctest.builddir) / 'test' / 'out.expected') as f:
        out_expected = f.read()
        assert out == out_expected


def test_new_file_with_out_of_date_cache(fctest, fcfont):
    fctest.setup()
    fctest.install_font(fcfont.fonts[0], '.')
    for ret, stdout, stderr in fctest.run_cache([fctest.fontdir.name]):
        assert ret == 0, stderr
    time.sleep(1)
    fctest.install_font(fcfont.fonts[1], 'a')
    l = []
    for ret, stdout, stderr in fctest.run_list(['-', 'family', 'pixelsize']):
        assert ret == 0, stderr
        l += sorted(stdout.splitlines())
    l += ['=']
    for ret, stdout, stderr in fctest.run_list(['-', 'family', 'pixelsize']):
        assert ret == 0, stderr
        l += sorted(stdout.splitlines())
    l += ['=']
    for ret, stdout, stderr in fctest.run_list(['-', 'family', 'pixelsize']):
        assert ret == 0, stderr
        l += sorted(stdout.splitlines())
    l += ['']
    out = '\n'.join(l)
    with open(Path(fctest.builddir) / 'test' / 'out.expected') as f:
        out_expected = f.read()
        assert out == out_expected


def test_keep_mtime(fctest, fcfont):
    fctest.setup()
    fctest.install_font(fcfont.fonts, '.', 0)
    fontdir = Path(fctest.fontdir.name)
    before = fontdir.stat().st_mtime
    time.sleep(1)
    for ret, stdout, stderr in fctest.run_cache([fctest.fontdir.name]):
        assert ret == 0, stderr
    after = fontdir.stat().st_mtime
    assert before == after, f'mtime {before} was changed to {after}'


def test_multiple_caches(fctest, fcfont):
    extraconffile = NamedTemporaryFile(prefix='fontconfig.',
                                       suffix='.extra.conf',
                                       mode='w',
                                       delete_on_close=False)
    fctest._extra.append(f'<include ignore_missing="yes">{fctest.convert_path(extraconffile.name)}</include>')

    # Set up for generating original caches
    fctest.setup()
    origepoch = epoch = os.getenv('SOURCE_DATE_EPOCH')
    if epoch:
        # epoch 0 has special meaning. increase to avoid epoch 0
        epoch = int(epoch) + 1
    fctest.install_font(fcfont.fonts, '.', epoch)
    if epoch:
        fctest._env['SOURCE_DATE_EPOCH'] = str(epoch)
    for ret, stdout, stderr in fctest.run_cache([fctest.convert_path(fctest.fontdir.name)]):
        assert ret == 0, stderr
    time.sleep(1)

    cache_files1 = [f.stat() for f in fctest.cache_files()]
    assert len(cache_files1) == 1, cache_files1

    # Set up for modified caches
    oldcachedir = fctest.cachedir
    oldconffile = fctest._conffile
    newcachedir = TemporaryDirectory(prefix='fontconfig.',
                                     suffix='.newcachedir')
    newconffile = NamedTemporaryFile(prefix='fontconfig.',
                                     suffix='.new.conf',
                                     mode='w',
                                     delete_on_close=False)
    fctest._cachedir = newcachedir
    fctest._conffile = newconffile
    fctest.setup()

    extraconffile.write(f'''
<fontconfig>
  <match target="scan">
    <test name="file"><string>{fctest.convert_path(fctest.fontdir.name)}/4x6.pcf</string></test>
    <edit name="pixelsize"><int>8</int></edit>
  </match>
</fontconfig>''')
    extraconffile.close()
    if epoch:
        fctest._env['SOURCE_DATE_EPOCH'] = str(epoch + 1)
    for ret, stdout, stderr in fctest.run_cache([fctest.convert_path(fctest.fontdir.name)]):
        assert ret == 0, stderr
    if epoch:
        fctest._env['SOURCE_DATE_EPOCH'] = origepoch
    cache_files2 = [f.stat() for f in fctest.cache_files()]
    assert len(cache_files2) == 1, cache_files2
    # Make sure if 1 and 2 is different
    assert cache_files1 != cache_files2

    ## Set up for mixed caches
    mixedconffile = NamedTemporaryFile(prefix='fontconfig.',
                                       suffix='.mixed.conf',
                                       mode='w',
                                       delete_on_close=False)
    fctest._cachedir = oldcachedir
    fctest._conffile = mixedconffile
    fctest._extra.append(f'<cachedir>{fctest.convert_path(newcachedir.name)}</cachedir>')
    fctest.setup()
    l = []
    for ret, stdout, stderr in fctest.run_list(['-', 'family', 'pixelsize']):
        assert ret == 0, stderr
        l += sorted(stdout.splitlines())
    l += ['=']
    for ret, stdout, stderr in fctest.run_list(['-', 'family', 'pixelsize']):
        assert ret == 0, stderr
        l += sorted(stdout.splitlines())
    l += ['=']
    for ret, stdout, stderr in fctest.run_list(['-', 'family', 'pixelsize']):
        assert ret == 0, stderr
        l += sorted(stdout.splitlines())
    l += ['']
    out = '\n'.join(l)
    with open(Path(fctest.builddir) / 'test' / 'out.expected') as f:
        s = f.read()
        out_expected = s.replace('pixelsize=6', 'pixelsize=8')
        assert out == out_expected

    del oldconffile


@pytest.mark.skipif(not not os.getenv('EXEEXT'), reason='not working on Win32')
def test_xdg_cache_home(fctest, fcfont):
    fctest._env['XDG_CACHE_HOME'] = ''
    old_home = os.getenv('HOME')
    new_home = TemporaryDirectory(prefix='fontconfig.',
                                  suffix='.home')
    fctest._env['HOME'] = new_home.name

    fctest.install_font(fcfont.fonts, '.')

    def custom_config(self):
        return f'<fontconfig><dir>{fctest.fontdir.name}</dir><cachedir prefix="xdg">fontconfig</cachedir></fontconfig>'

    fctest.config = types.MethodType(custom_config, fctest)
    fctest.setup()
    for ret, stdout, stderr in fctest.run_cache([fctest.fontdir.name]):
        assert ret == 0, stderr

    phome = Path(new_home.name)
    assert (phome / '.cache').exists()
    assert (phome / '.cache' / 'fontconfig').exists()
    cache_files = [f.name for f in (phome / '.cache' / 'fontconfig').glob('*cache*')]
    assert len(cache_files) == 1



# Copyright (C) 2025 fontconfig Authors
# SPDX-License-Identifier: HPND

from fctest import FcTest, FcExternalTestFont, FcTestFont
from pathlib import Path
from tempfile import NamedTemporaryFile
import pytest


@pytest.fixture
def fctest():
    return FcTest()


@pytest.fixture
def fcfont():
    return FcTestFont()


@pytest.mark.parametrize("font_file", FcExternalTestFont().fonts)
def test_genconf(fctest, fcfont, font_file):
    font_path = Path(font_file)

    if not font_path.exists():
        pytest.skip(f"Font file not found: {font_file}")

    extraconffile = NamedTemporaryFile(prefix='fontconfig.',
                                       suffix='.extra.conf',
                                       mode='w',
                                       delete_on_close=False)
    fctest._extra.append(f'<include ignore_missing="yes">{fctest.convert_path(extraconffile.name)}</include>')
    fctest.setup()
    fctest.install_font(fcfont.fonts, '.')
    fctest.install_font(font_file, '.')

    f = g = fmt = None
    for ret, stdout, stderr in fctest.run_query(['-f', '%{family}:%{genericfamily}\n', font_file]):
        assert ret == 0, stderr
        for l in stdout.strip().splitlines():
            f, g = l.split(':')
            assert f, stdout
            if not g or g == '0':
                # fallback if missing
                g = '2'
                fmt = ':family=sans-serif'
            else:
                fmt = f':genericfamily={g}'

    for ret, stdout, stderr in fctest.run_genconf(['-g', g, '-f', f, '-o', extraconffile.name, font_file]):
        assert ret == 0, stderr

    for ret, stdout, stderr in fctest.run_match(['-f', '%{file}\n', fmt]):
        assert ret == 0, stderr
        if Path(stdout.strip().splitlines()[0]).name != font_path.name:
            assert ret == 0, stderr

    for ret, stdout, stderr in fctest.run_scan(['-f', '%{genericfamily}\n', font_file]):
        assert ret == 0, stderr
        for l in stdout.strip().splitlines():
            assert g == l, stdout



# Copyright (C) 2025 fontconfig Authors
# SPDX-License-Identifier: HPND

from fctest import FcTest, FcTestFont
from operator import attrgetter
from pathlib import Path
from tempfile import TemporaryDirectory
import os
import pytest
import re
import shutil
import tempfile
import time
import types


@pytest.fixture
def fctest():
    return FcTest()


def dict_conf_with_json():
    srcdir = os.getenv('srcdir', Path(__file__).parent.parent)
    ret = {}
    for fn in (Path(srcdir) / 'conf.d').glob('*.conf'):
        json = Path(srcdir) / 'test' / ('test-' + fn.stem + '.json')
        if json.exists():
            ret[str(fn)] = str(json)

    return ret


def list_json():
    srcdir = os.getenv('srcdir', Path(__file__).parent.parent)
    pairs = list(dict_conf_with_json().values())
    FcTest().logger.info(pairs)
    ret = []
    for fn in (Path(srcdir) / 'test').glob('test-*.json'):
        if str(fn) not in pairs:
            ret.append(str(fn))

    return ret


@pytest.mark.parametrize(
    'conf, json',
    [(k, v) for k, v in dict_conf_with_json().items()],
    ids=lambda x: Path(x).name)
def test_pair_of_conf_and_json(fctest, conf, json):
    testexe = Path(fctest.builddir) / 'test' / ('test-conf' + fctest._exeext)
    if not testexe.exists():
        testexe = Path(fctest.builddir) / 'test' / ('test_conf' + fctest._exeext)
        if not testexe.exists():
            pytest.skip('No test executable. maybe missing json-c dependency?')

    for ret, stdout, stderr in fctest.run(testexe, [conf, json]):
        assert ret == 0, f'stdout:\n{stdout}\nstderr:\n{stderr}'
        fctest.logger.info(stdout)


@pytest.mark.parametrize('json', list_json(), ids=lambda x: Path(x).name)
def test_json(fctest, json):
    testexe = Path(fctest.builddir) / 'test' / ('test-conf' + fctest._exeext)
    if not testexe.exists():
        testexe = Path(fctest.builddir) / 'test' / ('test_conf' + fctest._exeext)
        if not testexe.exists():
            pytest.skip('No test executable. maybe missing json-c dependency?')
    harmlessconf = str(Path(fctest.srcdir) / 'conf.d' / '10-autohint.conf')

    for ret, stdout, stderr in fctest.run(testexe, [harmlessconf, json]):
        assert ret == 0, f'stdout:\n{stdout}\nstderr:\n{stderr}'
        fctest.logger.info(stdout)
